from typing import List, Sequence
from fastapi import Depends, HTTPException, Request, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
import argon2
from bmsdna.lakeapi.core.config import BasicConfig, Configs, UserConfig
from bmsdna.lakeapi.core.yaml import get_yaml
import inspect
from aiocache import cached, Cache
from aiocache.serializers import PickleSerializer
from bmsdna.lakeapi.core.env import CACHE_EXPIRATION_TIME_SECONDS

cache = cached(
    ttl=CACHE_EXPIRATION_TIME_SECONDS * 10000,
    cache=Cache.MEMORY,
    serializer=PickleSerializer(),
)

security = HTTPBasic()


userhashmap: dict[str, str] | None = None


@cache
async def is_correct(hash: str, pwd_str: str):
    ph = argon2.PasswordHasher()
    return ph.verify(hash.encode("utf-8"), pwd_str.encode("utf-8"))


async def get_username(req: Request, basic_config: BasicConfig, users: Sequence[UserConfig]):
    if req.query_params.get("token") and basic_config.token_jwt_secret is not None:
        import jwt

        token = req.query_params["token"]
        dt = jwt.decode(token, basic_config.token_jwt_secret, algorithms=["HS256"])
        if dt["path"] == req.url.path or dt["host"] == req.url.hostname:
            return dt["username"]
    credentials = await HTTPBasic(auto_error=True)(req)
    assert credentials is not None
    global userhashmap
    userhashmap = userhashmap or {
        ud["name"].casefold(): ud["passwordhash"] for ud in users if ud["name"]
    }  # pay attention not to include an empty user by accident

    if not credentials.username.casefold() in userhashmap.keys():
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect email or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    pwd_str = credentials.password
    hash = userhashmap[credentials.username.casefold()]

    is_correct_password = is_correct(hash, pwd_str)
    if not isinstance(is_correct_password, bool):
        is_correct_password = await is_correct_password
    if not is_correct_password:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect email or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    return credentials.username



from deltalake import DeltaTable

import pyarrow as pa
from typing import List, Tuple, Any, Union, TYPE_CHECKING
from bmsdna.lakeapi.core.types import FileTypes
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData

if TYPE_CHECKING:
    import datafusion  # lazy import it on runtime
import pyarrow.dataset
import pypika.queries
import pypika


class DatafusionDBResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[pypika.queries.QueryBuilder, str],
        session: "datafusion.SessionContext",
    ) -> None:
        super().__init__()
        self.original_sql = original_sql
        self.session = session
        self._arrow_schema = None
        self._df = None

    def columns(self):
        return self.arrow_schema().column_names

    def query_builder(self) -> pypika.queries.QueryBuilder:
        return pypika.Query.from_(self.original_sql)

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = (
            self.original_sql.limit(0).get_sql()
            if not isinstance(self.original_sql, str)
            else "SELECT * FROM (" + self.original_sql + ") s LIMIT 0 "
        )
        self._arrow_schema = self.session.sql(query).to_arrow_table()
        return self._arrow_schema

    @property
    def df(self):
        if self._df is None:
            self._df = self.session.sql(
                self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
            )
        return self._df

    def to_pandas(self):
        return self.df.to_pandas()

    def to_arrow_table(self):
        return self.df.to_arrow_table()

    def to_arrow_recordbatch(self, chunk_size: int = 10000):
        return self.df.to_arrow_table().to_reader(max_chunksize=chunk_size)


class DatafusionDbExecutionContextBase(ExecutionContext):
    def __init__(self, session: "datafusion.SessionContext"):
        super().__init__()
        self.session = session

    def register_arrow(
        self,
        name: str,
        ds: Union[pyarrow.dataset.Dataset, pyarrow.Table, pyarrow.dataset.FileSystemDataset],
    ):
        if isinstance(ds, pyarrow.dataset.Dataset):
            self.session.deregister_table(name)
            self.session.register_dataset(name, ds)
        elif isinstance(ds, pyarrow.dataset.FileSystemDataset):
            self.session.deregister_table(name)
            self.session.register_dataset(name, ds)
        elif isinstance(ds, pyarrow.Table):
            self.session.deregister_table(name)
            self.session.register_table(name, ds)
        else:
            raise ValueError("not supported")

    def close(self):
        pass

    def execute_sql(
        self,
        sql: Union[
            pypika.queries.QueryBuilder,
            str,
        ],
    ) -> DatafusionDBResultData:
        return DatafusionDBResultData(sql, session=self.session)


class DatafusionDbExecutionContext(DatafusionDbExecutionContextBase):
    def __init__(self):
        import datafusion

        runtime = datafusion.RuntimeConfig().with_disk_manager_os()
        config = (
            datafusion.SessionConfig()
            .set("datafusion.execution.parquet.pushdown_filters", "true")
            .set("datafusion.execution.collect_statistics", "true")
            .set("datafusion.execution.parquet.enable_page_index", "true")
            .set("datafusion.sql_parser.enable_ident_normalization", "false")
        )
        ctx = datafusion.SessionContext(config, runtime)

        super().__init__(ctx)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



from datetime import datetime
from deltalake import DeltaTable

import pyarrow as pa
from typing import List, Optional, Tuple, Any, Union
from bmsdna.lakeapi.core.types import FileTypes
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData
import duckdb
import pyarrow.dataset
import pypika.queries
import pypika.terms
import pypika
import os
from datetime import datetime, timezone
from bmsdna.lakeapi.core.config import SearchConfig

ENABLE_COPY_TO = os.getenv("ENABLE_COPY_TO", "0") == "1"


class DuckDBResultData(ResultData):
    def __init__(
        self,
        original_sql: Union[pypika.queries.QueryBuilder, str],
        con: duckdb.DuckDBPyConnection,
    ) -> None:
        super().__init__()
        self.original_sql = original_sql
        self.con = con
        self._arrow_schema = None
        self._df = None

    def columns(self):
        return self.arrow_schema().names

    def query_builder(self) -> pypika.queries.QueryBuilder:
        return pypika.Query.from_(self.original_sql)

    def arrow_schema(self) -> pa.Schema:
        if self._arrow_schema is not None:
            return self._arrow_schema
        query = (
            self.original_sql.limit(0).get_sql()
            if not isinstance(self.original_sql, str)
            else "SELECT * FROM (" + self.original_sql + ") s LIMIT 0 "
        )
        self._arrow_schema = self.con.execute(query).arrow().schema
        return self._arrow_schema

    @property
    def df(self):
        if self._df is None:
            query = self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
            self._df = self.con.execute(query)
        return self._df

    def to_pandas(self):
        return self.df.df()

    def to_arrow_table(self):
        return self.df.arrow()

    def to_arrow_recordbatch(self, chunk_size: int = 10000):
        return self.df.fetch_record_batch(chunk_size)

    def write_parquet(self, file_name: str):
        if not ENABLE_COPY_TO:
            return super().write_parquet(file_name)
        query = self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
        full_query = f"COPY ({query}) TO '{file_name}' (FORMAT PARQUET,use_tmp_file False, ROW_GROUP_SIZE 10000)"
        self.con.execute(full_query)

    def write_nd_json(self, file_name: str):
        if not ENABLE_COPY_TO:
            return super().write_nd_json(file_name)
        query = self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
        full_query = f"COPY ({query}) TO '{file_name}' (FORMAT JSON)"
        self.con.execute(full_query)

    def write_csv(self, file_name: str, *, separator: str):
        if not ENABLE_COPY_TO:
            return super().write_csv(file_name, separator=separator)
        query = self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
        full_query = f"COPY ({query}) TO '{file_name}' (FORMAT CSV, delim '{separator}', header True)"
        self.con.execute(full_query)

    def write_json(self, file_name: str):
        if not ENABLE_COPY_TO:
            return super().write_json(file_name)
        query = self.original_sql if isinstance(self.original_sql, str) else self.original_sql.get_sql()
        full_query = f"COPY ({query}) TO '{file_name}' (FORMAT JSON, Array True)"
        self.con.execute(full_query)


class Match25Term(pypika.terms.Term):
    def __init__(
        self,
        source_view: str,
        field: pypika.terms.Term,
        search_text: str,
        fields: Optional[str],
        alias: Optional[str] = None,
    ):
        super().__init__()
        self.source_view = source_view
        self.field = field
        self.search_text = search_text
        self.fields = fields
        self.alias = alias

    def get_sql(self, **kwargs):
        search_text_const = pypika.terms.Term.wrap_constant(self.search_text)
        assert isinstance(search_text_const, pypika.terms.Term)
        search_txt = search_text_const.get_sql()
        fields_const = pypika.terms.Term.wrap_constant(self.fields or "")
        assert isinstance(fields_const, pypika.terms.Term)
        field_or_not = ", fields := " + fields_const.get_sql() if self.fields is not None else ""
        sql = f"fts_main_{self.source_view}.match_bm25({self.field.get_sql()}, {search_txt}{field_or_not})"
        if self.alias is not None:
            sql += " AS " + self.alias
        return sql


class DuckDbExecutionContextBase(ExecutionContext):
    def __init__(self, con: duckdb.DuckDBPyConnection):
        super().__init__()
        self.con = con
        self.res_con = None
        self.persistance_file_name = None

    def register_arrow(self, name: str, ds: Union[pyarrow.dataset.Dataset, pyarrow.Table]):
        # self.con.from_arrow(ds).create_view(name, replace=True)
        self.con.register(name, ds)

    def close(self):
        pass

    def execute_sql(
        self,
        sql: Union[
            pypika.queries.QueryBuilder,
            str,
        ],
    ) -> DuckDBResultData:
        if self.persistance_file_name is not None:
            self.res_con = duckdb.connect(self.persistance_file_name, read_only=True)
            return DuckDBResultData(sql, self.res_con)
        return DuckDBResultData(sql, con=self.con)

    def search_score_function(
        self,
        source_view: str,
        search_text: str,
        search_config: SearchConfig,
        alias: Optional[str],
    ):
        fields = ",".join(search_config.columns)
        return Match25Term(source_view, pypika.queries.Field("__search_id"), search_text, fields)

    def init_search(
        self,
        source_view: str,
        search_configs: list[SearchConfig],
    ):
        modified_date = self.modified_dates[source_view]
        persistence_name = source_view
        search_columns = []
        for cfg in search_configs:
            search_columns = search_columns + cfg.columns
        scc = ", ".join([f"{sc}" for sc in search_columns])
        pk_name = "__search_id"
        real_query = f"CREATE TABLE search_con.{persistence_name} AS SELECT ROW_NUMBER() OVER () AS __search_id, * FROM {source_view} s"
        persitance_path = os.getenv("TEMP", "/tmp/")
        persistance_file_name = os.path.join(persitance_path, persistence_name + ".duckdb")

        if (
            not os.path.exists(persistance_file_name)
            or datetime.fromtimestamp(os.path.getmtime(persistance_file_name), tz=timezone.utc)
            < modified_date.astimezone(timezone.utc)
            or datetime.fromtimestamp(os.path.getmtime(persistance_file_name), tz=timezone.utc)
            < datetime.fromisoformat("2023-05-19T00:00Z")  # before duckdb upgrade
        ):
            persistance_file_name_temp = persistance_file_name + "_temp"
            if os.path.exists(persistance_file_name_temp):
                os.remove(persistance_file_name_temp)
            search_con = duckdb.connect(persistance_file_name_temp, read_only=False)
            search_con.commit()  # create empty duck file
            search_con.close()
            self.con.execute(f"ATTACH '{persistance_file_name_temp}' AS search_con")
            self.con.execute(real_query)
            self.con.execute("DETACH search_con")
            search_con = duckdb.connect(persistance_file_name_temp, read_only=False)
            scc = ", ".join([f"'{sc}'" for sc in search_columns])
            search_con.execute(f"PRAGMA create_fts_index('{persistence_name}', '{pk_name}', {scc})")
            search_con.close()
            if os.path.exists(persistance_file_name):
                os.remove(persistance_file_name)
            os.rename(persistance_file_name_temp, persistance_file_name)
        self.persistance_file_name = persistance_file_name

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.res_con:
            self.res_con.__exit__(*args)


class DuckDbExecutionContext(DuckDbExecutionContextBase):
    def __init__(self):
        super().__init__(duckdb.connect())

    def __enter__(self):
        super().__enter__()
        self.con.__enter__()
        self.con.execute("SET memory_limit='200MB'")
        self.con.execute("SET threads =2")
        self.con.execute("SET enable_object_cache=true")
        return self

    def __exit__(self, *args):
        super().__exit__(*args)
        self.con.__exit__(*args)

    def close(self):
        self.con.close()



from datetime import datetime
from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData


import pyarrow as pa
from typing import List, Optional, Tuple, Union, cast, TYPE_CHECKING, Any
import threading
from bmsdna.lakeapi.core.types import FileTypes
import pyarrow.dataset
import pypika.queries
import pypika
from uuid import uuid4

if TYPE_CHECKING:
    import polars  # we want to lazy import in case we one day no longer rely on polars if only duckdb is needed


class PolarsResultData(ResultData):
    def __init__(self, df: "Union[polars.DataFrame, polars.LazyFrame]", sql_context: "polars.SQLContext"):
        self.df = df
        self.random_name = "tbl_" + str(uuid4())
        self.registred_df = False
        self.sql_context = sql_context

    def columns(self):
        return self.df.columns

    def query_builder(self) -> pypika.queries.QueryBuilder:
        import polars as pl
        if not self.registred_df:
            if isinstance(self.df, pl.DataFrame):
                self.df = self.df.lazy()

            self.sql_context.register(self.random_name, cast(pl.LazyFrame, self.df))
            self.registred_df = True
        return pypika.Query.from_(self.random_name)

    def _to_arrow_type(self, t: "polars.PolarsDataType"):
        import polars as pl
        from polars.datatypes.convert import DataTypeMappings, py_type_to_arrow_type

        if isinstance(t, pl.Struct):
            return pa.struct({f.name: self._to_arrow_type(f.dtype) for f in t.fields})
        if isinstance(t, pl.List):
            return pa.list_(self._to_arrow_type(t.inner) if t.inner else pa.string)

        if t in DataTypeMappings.PY_TYPE_TO_ARROW_TYPE:
            return py_type_to_arrow_type(t)
        raise ValueError("not supported")

    def arrow_schema(self) -> pa.Schema:
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            return pa.Schema([pa.field(k, self._to_arrow_type(v)) for k, v in self.df.schema.items()])
        else:
            return self.df.limit(0).to_arrow().schema

    def to_pandas(self):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect(streaming=True)
        return self.df.to_pandas()

    def to_arrow_table(self):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect()
        assert isinstance(self.df, pl.DataFrame)
        return self.df.to_arrow()

    def to_arrow_recordbatch(self, chunk_size: int = 10000):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect()
        assert isinstance(self.df, pl.DataFrame)
        pat = self.df.to_arrow()
        return pat.to_reader(max_chunksize=chunk_size)

    def write_parquet(self, file_name: str):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect(streaming=True)
        self.df.write_parquet(file_name, use_pyarrow=True)

    def write_json(self, file_name: str):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect(streaming=True)
        self.df.write_json(file_name, pretty=False, row_oriented=True)

    def write_nd_json(self, file_name: str):
        import polars as pl

        if isinstance(self.df, pl.LazyFrame):
            self.df = self.df.collect(streaming=True)
        self.df.write_ndjson(file_name)


class PolarsExecutionContext(ExecutionContext):
    def __init__(self, sql_context: "Optional[polars.SQLContext]" = None):
        super().__init__()
        import polars as pl

        self.sql_context = sql_context or pl.SQLContext()

    def register_arrow(self, name: str, ds: Union[pyarrow.dataset.Dataset, pyarrow.Table]):
        import polars as pl

        ds = pl.scan_pyarrow_dataset(ds) if isinstance(ds, pyarrow.dataset.Dataset) else pl.from_arrow(ds)

        if isinstance(ds, pl.DataFrame):
            ds = ds.lazy()

        ds = pl.scan_ds(ds)
        self.sql_context.register(name, ds)

    def close(self):
        pass

    def register_dataframe(
        self,
        name: str,
        uri: str,
        file_type: FileTypes,
        partitions: Optional[List[Tuple[str, str, Any]]],
    ):
        import polars as pl

        self.modified_dates[name] = self.get_modified_date(uri, file_type)
        match file_type:
            case "delta":
                from bmsdna.lakeapi.polars_extensions.delta import scan_delta2

                df = pl.scan_delta2(  # type: ignore
                    uri,
                    pyarrow_options={"partitions": partitions},
                )
            case "parquet_withdecimal":  # differentiation no longer needed
                df = pl.scan_parquet(uri)
            case "parquet":
                df = pl.scan_parquet(uri)
            case "arrow":
                df = pl.scan_ipc(uri)
            case "avro":
                df = cast(pl.LazyFrame, pl.read_avro(uri))
            case "csv":
                df = pl.scan_csv(uri)
            case "json":
                df = cast(pl.LazyFrame, pl.read_json(uri))
            case "ndjson":
                df = pl.scan_ndjson(uri)
            case _:
                raise Exception(f"Not supported file type {file_type}")
        self.sql_context.register(name, df)

    def execute_sql(
        self,
        sql: Union[
            pypika.queries.QueryBuilder,
            str,
        ],
    ) -> PolarsResultData:
        import polars as pl

        df = self.sql_context.execute(sql.get_sql() if not isinstance(sql, str) else sql)
        if isinstance(df, pl.LazyFrame):
            df = df.collect()
        return PolarsResultData(df, self.sql_context)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


local = threading.local()
local.exec_context = None


def get_polars_context() -> ExecutionContext:
    if local.exec_context is None:
        local.exec_context = PolarsExecutionContext()
    return local.exec_context



from typing import Literal
from datetime import datetime, date, time, timedelta
from decimal import Decimal
from typing import Type
from polars.datatypes.convert import DataTypeMappings


Engines = Literal["duckdb", "polars", "datafusion"]


FileTypes = Literal[
    "delta",
    "parquet",
    "arrow",
    "arrow-stream",
    "avro",
    "csv",
    "json",
    "ndjson",
    "parquet_withdecimal",
]
OutputFileType = Literal[
    "json",
    "ndjson",
    "parquet",
    "csv",
    "csv4excel",
    "xlsx",
    "feather",
    "html",
    "scsv",
    "xml",
    "ipc",
    "arrow",
    "arrow-stream",
]
OperatorType = Literal[
    "<",
    "=",
    ">",
    ">=",
    "<=",
    "<>",
    "contains",
    "in",
    "not contains",
    "not in",
    "not null",
    "null",
    "between",
    "not between",
]
DeltaOperatorTypes = Literal["<", "=", ">", ">=", "<=", "in", "not in"]
PolaryTypeFunction = Literal[
    "sum",
    "count",
    "mean",
    "median",
    "min",
    "max",
    "first",
    "std",
    "n_unique",
    "distinct",
    "mode",
    "null_count",
]

CONFIG_DTYPE_MAP: dict[str, Type] = {
    "struct": dict,
    "string": str,
    "integer": int,
    "float": float,
    "array": list,
    "double": Decimal,
    "timesamp": datetime,
    "date": date,
    "time": time,
    "duration": timedelta,
    "str": str,
    "int": int,
    "bool": bool,
    "boolean": bool,
}


DTYPE_MAP = {
    "struct": dict,
    "string": str,
    "integer": int,
    "float": float,
    "array": list,
    "double": Decimal,
    "timesamp": datetime,
    "date": date,
    "time": time,
}



import io
import json
import mimetypes
import os
from enum import Enum
from typing import Union
from uuid import uuid4

import anyio
import pyarrow as pa
from fastapi import BackgroundTasks
from starlette.datastructures import URL
from starlette.responses import FileResponse

from bmsdna.lakeapi.context.df_base import ExecutionContext, ResultData
from bmsdna.lakeapi.core.config import BasicConfig
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.types import OutputFileType

logger = get_logger(__name__)


class OutputFormats(Enum):
    AVRO = 1
    CSV = 2
    CSV4EXCEL = 12
    SEMI_CSV = 11
    XLSX = 3
    HTML = 4
    ARROW_IPC = 5
    ND_JSON = 6
    PARQUET = 7
    JSON = 8
    XML = 9
    ORC = 10
    IQY = 13
    ARROW_STREAM = 14


def parse_format(accept: Union[str, OutputFileType]) -> tuple[OutputFormats, str]:
    realaccept = accept.split(";")[0].strip().lower()
    if realaccept == "application/avro" or realaccept == "avro":
        return (OutputFormats.AVRO, ".avro")
    elif realaccept == "text/csv" or realaccept == "csv":
        return (OutputFormats.CSV, ".csv")
    elif realaccept == "text/csv+semicolon" or realaccept == "scsv":
        return (OutputFormats.SEMI_CSV, ".csv")
    elif realaccept == "x.text/csv+excel" or realaccept == "csv4excel":
        return (OutputFormats.CSV4EXCEL, ".csv")
    elif realaccept == "application/xml" or realaccept == "xml":
        return (OutputFormats.XML, ".xml")
    elif realaccept == "orc":
        return (OutputFormats.ORC, ".orc")
    elif realaccept == mimetypes.guess_type("file.xlsx") or realaccept == "xlsx":
        return (OutputFormats.XLSX, ".xlsx")
    elif realaccept == "text/html" or realaccept == "html":
        return (OutputFormats.HTML, ".html")
    elif realaccept == "x-application/iqy" or realaccept == "iqy":
        return (OutputFormats.IQY, ".iqy")

    elif realaccept == "application/vnd.apache.arrow.stream" or realaccept == "arrow-stream":
        return (OutputFormats.ARROW_STREAM, "")

    elif (
        realaccept == "application/x-arrow"
        or realaccept == "application/x-arrow"
        or realaccept == "application/vnd.apache.arrow.file"
        or realaccept == "arrow"
        or realaccept == "feather"
        or realaccept == "ipc"
    ):
        return (OutputFormats.ARROW_IPC, ".arrow")
    elif realaccept == "application/json+newline" or realaccept == "application/jsonl" or realaccept == "ndjson":
        return (OutputFormats.ND_JSON, ".ndjson")
    elif realaccept == "application/parquet" or realaccept == "parquet":
        return (OutputFormats.PARQUET, ".parquet")
    else:
        return (OutputFormats.JSON, ".json")


def write_frame(
    url: URL, current_user: str, content: ResultData, format: OutputFormats, out: str, basic_config: BasicConfig
):
    if format == OutputFormats.IQY:
        import jwt

        import env

        query = url.query.replace("format=iqy", "").rstrip("&")
        query = query + ("?" if not query else "&") + "format=json"
        token = jwt.encode(
            {"host": url.hostname, "path": url.path, "username": current_user},
            basic_config.token_jwt_secret,
            algorithm="HS256",
        )
        query += f"&token={token}"
        host_and_port = (url.hostname or "localhost") + (":" + str(url.port) if url.port and url.port != 443 else "")
        strdt = f"""WEB
1
{url.scheme}://{host_and_port}{url.path}?{query}

Selection=1
Formatting=None
PreFormattedTextToColumns=True
ConsecutiveDelimitersAsOne=False
SingleBlockTextImport=False
DisableDateRecognition=False
DisableRedirections=False"""
        if isinstance(out, str):
            with open(out, "w") as f:
                f.write(strdt)
        else:
            out.write(strdt.encode("utf-8"))
        return

    if format == OutputFormats.AVRO:
        import polars as pl

        ds = pl.from_arrow(content.to_arrow_recordbatch())
        assert isinstance(ds, pl.DataFrame)
        ds.write_avro(out)
    elif format == OutputFormats.CSV:
        content.write_csv(out, separator=",")
    elif format == OutputFormats.SEMI_CSV:
        content.write_csv(out, separator=";")
    elif format == OutputFormats.CSV4EXCEL:  # need to write sep=, on first line
        if isinstance(out, str):
            content.write_csv(out + "_u8", separator=",")  # type: ignore
            with open(
                out, mode="wb"
            ) as f:  # excel wants utf-16le which polars does not support. therefore we need reencoding
                f.write(b"sep=,\n")  # add utf-8 bom at beginning
                with open(out + "_u8", mode="r", encoding="utf-8") as c8:
                    line = c8.readline()
                    while line != "":
                        f.write(line.encode("utf-16le"))
                        line = c8.readline()
                os.remove(out + "_u8")

        else:
            out.write(b"sep=,\r\n")  # not used, and would break special characters
            content.write_csv(out, separator=",")
    elif format == OutputFormats.XLSX:
        import polars as pl

        ds = pl.from_arrow(content.to_arrow_table())
        assert isinstance(ds, pl.DataFrame)
        ds.write_excel(out, autofit=True)
    elif format == OutputFormats.HTML:
        content.to_pandas().to_html(out, index=False)

    elif format == OutputFormats.XML:
        content.to_pandas().to_xml(out, index=False, parser="etree")

    elif format in (OutputFormats.ARROW_IPC, OutputFormats.ARROW_STREAM):
        with content.to_arrow_recordbatch() as batches:
            with pa.OSFile(out, "wb") as sink:
                with pa.ipc.new_file(sink, batches.schema) as writer:
                    for batch in batches:
                        writer.write(batch)

    elif format == OutputFormats.ND_JSON:
        content.write_nd_json(out)
    elif format == OutputFormats.PARQUET:
        content.write_parquet(out)
    else:
        content.write_json(out)


async def create_response(
    current_user_name: str,
    url: URL,
    accept: str,
    content: ResultData,
    context: ExecutionContext,
    basic_config: BasicConfig,
):
    headers = {}

    format = parse_format(accept)

    format, extension = parse_format(accept)
    content_dispositiont_type = "attachment"
    filename = "file" + extension
    if format in [
        OutputFormats.JSON,
        OutputFormats.ND_JSON,
        OutputFormats.CSV,
        OutputFormats.SEMI_CSV,
        OutputFormats.CSV4EXCEL,
    ]:
        content_dispositiont_type = "inline"
        filename = None
    path = os.path.join(basic_config.temp_folder_path, str(uuid4()) + extension)
    media_type = mimetypes.guess_type("file" + extension)[0]
    write_frame(
        current_user=current_user_name, url=url, content=content, format=format, out=path, basic_config=basic_config
    )

    if media_type is not None and format in [
        OutputFormats.JSON,
        OutputFormats.ND_JSON,
        OutputFormats.CSV,
        OutputFormats.SEMI_CSV,
        OutputFormats.CSV4EXCEL,
        OutputFormats.HTML,
        OutputFormats.XLSX,
        OutputFormats.XML,
    ]:
        media_type = media_type + ";charset=utf-8"
    tasks = BackgroundTasks()
    import asyncio

    async def remove():
        context.close()
        await anyio.sleep(3)

        os.remove(path)

    tasks.add_task(remove)

    return FileResponse(
        path=path,
        headers=headers,
        media_type=media_type,
        content_disposition_type=content_dispositiont_type,
        filename=filename,
        background=tasks,
    )



from fastapi import Depends, FastAPI

from bmsdna.lakeapi.api.api import init_lakeapi


def run_fastapi():
    app = FastAPI()
    sti = init_lakeapi(app)

    @app.get("/")
    async def root(username: str = Depends(sti.get_username)):
        return {"User": username}

    return app


app = run_fastapi()



import ast
import os
import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    Union,
    cast,
)
from typing_extensions import TypedDict, NotRequired, Required
from fastapi import APIRouter, Request

import yaml
from polars.type_aliases import JoinStrategy
from bmsdna.lakeapi.core.env import CACHE_EXPIRATION_TIME_SECONDS

from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.types import FileTypes, OperatorType, PolaryTypeFunction, Engines

logger = get_logger(__name__)


@dataclass(frozen=True)
class BasicConfig:
    username_retriever: Callable[[Request, "BasicConfig", "Sequence[UserConfig]"], str | Awaitable[str]]
    enable_sql_endpoint: bool
    temp_folder_path: str
    data_path: str
    token_jwt_secret: str | None  # None disables the token feature
    min_search_length: int


def get_default_config():
    from bmsdna.lakeapi.core.uservalidation import get_username

    return BasicConfig(
        username_retriever=get_username,
        enable_sql_endpoint=os.getenv("ENABLE_SQL_ENDPOINT", "0") == "1",
        temp_folder_path=os.getenv("TEMP", "/tmp"),
        data_path=os.environ.get("DATA_PATH", "data"),
        token_jwt_secret=os.getenv("JWT_SECRET", None),
        min_search_length=3,
    )


@dataclass
class Param:
    name: str
    combi: Optional[Optional[List[str]]] = None
    default: Optional[str] = None
    required: Optional[bool] = False
    operators: Optional[list[OperatorType]] = None
    colname: Optional[str] = None

    @property
    def real_default(self) -> str:
        self._real_default = ast.literal_eval(self.default) if self.default else None
        return cast(str, self._real_default)


@dataclass
class Filter:
    key: str
    operator: Literal[
        "=",
        ">",
        "<",
        ">=",
        "<=",
        "<>",
        "!=",
        "==",
        "between",
        "not between",
        "in",
        "not in",
    ]
    value: Union[Any, Tuple[Any, Any]]


@dataclass
class SortBy:
    by: str
    direction: Optional[Literal["asc", "desc"]] = "asc"


@dataclass
class Column:
    name: str
    alias: Optional[str] = None

    def __post_init__(self):
        self.alias = self.alias if self.alias else self.name


@dataclass
class GroupByExpConfig:
    # expression: str
    col: str = ""
    func: Optional[PolaryTypeFunction] = None
    alias: Optional[str] = None

    @classmethod
    def from_dict(cls, config: Dict):
        return cls(col=config["col"], func=config.get("func"), alias=config.get("alias"))


@dataclass
class GroupByConfig:
    by: List[str]
    expressions: List[GroupByExpConfig]


@dataclass
class Join:
    uri: str
    left_on: str
    right_on: str
    file_type: FileTypes = "delta"
    how: JoinStrategy = "inner"
    suffix: str = "_right"


@dataclass
class DataframeConfig:
    uri: str
    file_type: FileTypes = "delta"
    select: Optional[List[Column]] = None
    exclude: Optional[List[str]] = None
    groupby: Optional[GroupByConfig] = None
    sortby: Optional[List[SortBy]] = None
    joins: Optional[List[Join]] = None
    filters: Optional[List[Filter]] = None
    cache_expiration_time_seconds: Optional[int] = CACHE_EXPIRATION_TIME_SECONDS


@dataclass
class SearchConfig:
    name: str
    columns: List[str]


@dataclass
class Option:
    ...


@dataclass
class Config:
    name: str
    tag: str
    dataframe: DataframeConfig
    version: Optional[int] = 1
    api_method: Union[Literal["get", "post"], List[Literal["get", "post"]]] = "get"
    params: Optional[List[Union[Param, str]]] = None
    engine: Engines = "duckdb"
    timestamp: Optional[datetime] = None
    cache_expiration_time_seconds: Optional[int] = CACHE_EXPIRATION_TIME_SECONDS
    options: Optional[Union[Option, None]] = None
    allow_get_all_pages: Optional[bool] = False
    search: Optional[List[SearchConfig]] = None

    def __post_init__(self):
        self.version_str = (
            str(self.version) if str(self.version or 1).startswith("v") else "v" + str(self.version or 1)
        )
        self.route = "/api/" + self.version_str + "/" + self.tag + "/" + self.name

    def __repr__(self) -> str:
        kws = [f"{key}={value!r}" for key, value in self.__dict__.items()]
        return "{}({})".format(type(self).__name__, ", ".join(kws))

    def __str__(self) -> str:
        kws = [f"{key}={value!r}" for key, value in self.__dict__.items()]
        return "{}({})".format(type(self).__name__, ", ".join(kws))

    @classmethod
    def from_dict(cls, config: Dict, basic_config: BasicConfig, table_names: List[tuple[int, str, str]]) -> List["Config"]:
        name = config["name"]
        tag = config["tag"]
        version = config.get("version", 1)
        engine = config.get("engine", "duckdb")
        api_method = cast(Literal["post", "get"], config.get("api_method", "get"))
        params = config.get("params")
        dataframe = config["dataframe"]
        uri = dataframe["uri"]
        assert isinstance(uri, str)
        file_type = cast(FileTypes, dataframe.get("file_type", "delta") if dataframe else "delta")
        columns = dataframe.get("columns") if dataframe else None
        select = dataframe.get("select") if dataframe else None
        exclude = dataframe.get("exclude") if dataframe else None
        groupby = dataframe.get("groupby") if dataframe else None
        sortby = dataframe.get("sortby") if dataframe else None
        cache_expiration_time_seconds = (
            dataframe.get("cache_expiration_time_seconds", CACHE_EXPIRATION_TIME_SECONDS) if dataframe else None
        )
        orderby = dataframe.get("orderby") if dataframe else None
        joins = dataframe.get("joins") if dataframe else None
        search_config = [SearchConfig(**c) for c in config["search"]] if "search" in config else None
        logger.debug(config)

        _params: list[Param] = []
        if params:
            logger.debug(f"Parameter: {name} -> {params}")

            _params += [Param(name=param) if isinstance(param, str) else Param(**param) for param in params]

        if joins:
            logger.debug(f"Joins: {name} -> {joins}")
            joins = [Join(**join) for join in joins]

        if groupby:
            logger.debug(f"Groupby: {name} -> {groupby}")
            expressions = [GroupByExpConfig.from_dict(e) for e in groupby.get("expressions")]
            groupby = GroupByConfig(by=groupby.get("by"), expressions=expressions)

        sortby = orderby if orderby else sortby
        if sortby:
            logger.debug(f"SortBy: {name} -> {sortby}")
            sortby = [SortBy(by=s.get("by"), direction=s.get("direction")) for s in sortby]

        select = columns if columns else select

        if select:
            select = [Column(name=c.get("name"), alias=c.get("alias")) for c in select]

        if not engine:
            engine = "duckdb"

        if name == "*":
            assert uri.endswith("/*")
            folder = uri.rstrip("/*")
            root_folder = os.path.join(basic_config.data_path, folder)
            if not os.path.exists(root_folder):
                logger.warning("Path not existing: " + root_folder)
                return []
            else:
                ls = []
                for it in os.scandir(root_folder):
                    res_name = (version, tag, it.name)
                    if res_name not in table_names and (
                        (it.is_dir() and file_type == "delta") or (it.is_file() and file_type != "delta")
                    ):
                        dataframe = DataframeConfig(
                            uri=folder + "/" + it.name,
                            file_type=file_type,
                            select=select,
                            exclude=exclude,
                            groupby=groupby,
                            sortby=sortby,
                            joins=joins,
                            filters=None,
                            cache_expiration_time_seconds=cache_expiration_time_seconds,
                        )

                        ls.append(
                            cls(
                                name=it.name,
                                tag=tag,
                                engine=engine,  # type: ignore
                                version=version,
                                api_method=api_method,
                                search=search_config,
                                params=_params,  # type: ignore
                                allow_get_all_pages=config.get("allow_get_all_pages", False),
                                dataframe=dataframe,
                                cache_expiration_time_seconds=cache_expiration_time_seconds,
                            )
                        )
            return ls
        else:
            dataframe = DataframeConfig(
                uri=uri,
                file_type=file_type,
                select=select,
                exclude=exclude,
                groupby=groupby,
                sortby=sortby,
                joins=joins,
                filters=None,
                cache_expiration_time_seconds=cache_expiration_time_seconds,
            )

            return [
                cls(
                    name=name,
                    tag=tag,
                    version=version,
                    search=search_config,
                    engine=engine,  # type: ignore
                    api_method=api_method,
                    params=_params,  # type: ignore
                    allow_get_all_pages=config.get("allow_get_all_pages", False),
                    dataframe=dataframe,
                    cache_expiration_time_seconds=cache_expiration_time_seconds,
                )
            ]

    async def to_dict(self) -> dict:
        return self.__dict__


class HttpConfig(TypedDict):
    bind: NotRequired[str]
    max_requests: NotRequired[int]
    max_requests_jitter: NotRequired[int]
    timeout: NotRequired[int]
    graceful_timeout: NotRequired[int]


class UserConfig(TypedDict):
    name: Required[str]
    passwordhash: Required[str]


class AppConfig(TypedDict):
    tile: NotRequired[str]
    description: NotRequired[str]
    version: NotRequired[str]
    static_file_path: NotRequired[str]
    logo_path: NotRequired[str]


class YamlData(TypedDict):
    tables: list[Config]
    users: list[UserConfig]
    app: AppConfig


@dataclass(frozen=True)
class Configs:
    configs: Sequence[Config]
    users: Sequence[UserConfig]

    def get_config_by_route(self, route: str):
        for config in self.configs:
            if config.route == route:
                return config

    def __iter__(self):
        for i in self.configs:
            yield i

    @classmethod
    def from_yamls(
        cls,
        basic_config: BasicConfig,
        root: str = "config",
        exclude_internal: bool = True,
        internal_pattern: str = "_",
    ):
        if re.search(r"\.ya?ml$", root):
            files = [root]
        else:
            files = [
                os.path.join(root, file)
                for file in os.listdir(root)
                if not file.startswith(internal_pattern) or not exclude_internal
            ]

        tables = []
        users = []
        for file in files:
            if file.endswith(".yml"):
                with open(file, encoding="utf-8") as f:
                    y = yaml.safe_load(f)
                    yaml_data_tables = y.get("tables")
                    if yaml_data_tables:
                        tables += [y for y in yaml_data_tables if y.get("name")]
                    yaml_data_users = y.get("users")
                    if yaml_data_users:
                        users += yaml_data_users

        flat_map = lambda f, xs: [y for ys in xs for y in f(ys)]
        table_names = [(t.get("version", 1), t["tag"], t["name"]) for t in tables if t["name"] != "*"]
        configs = flat_map(lambda x: Config.from_dict(x, basic_config=basic_config, table_names=table_names), tables)
        return cls(configs, users)



import datetime
from typing import Any, Dict, List, Literal, Union, cast, Optional, Iterable
from bmsdna.lakeapi.context.df_base import ResultData
from aiocache import Cache, cached
from aiocache.serializers import PickleSerializer
from fastapi import Query
from pydantic import BaseModel, create_model
from bmsdna.lakeapi.context.df_base import ResultData
from bmsdna.lakeapi.core.config import Param, SearchConfig
from bmsdna.lakeapi.core.types import OperatorType
from bmsdna.lakeapi.core.env import CACHE_EXPIRATION_TIME_SECONDS
import pyarrow as pa

cache = cached(ttl=CACHE_EXPIRATION_TIME_SECONDS, cache=Cache.MEMORY, serializer=PickleSerializer())


def _make_model(v, name):
    if type(v) is dict:
        return create_model(name, **{k: _make_model(v, k) for k, v in v.items()}), ...
    return type(v), None


def make_model(v: Dict, name: str):
    return _make_model(v, name)[0]


empty_model = create_model("NoQuery")

_operator_postfix_map: dict[OperatorType, str] = {
    "<": "_lt",
    "<=": "_lte",
    "<>": "_ne",
    "=": "",
    ">": "_gt",
    ">=": "_gte",
    "contains": "_contains",
    "not contains": "_not_contains",
    "in": "_in",
    "not in": "_not_in",
    "not null": "_not_null",
    "null": "_null",
    "between": "_between",
    "not between": "_not_between",
}


@cache
async def get_param_def(queryname: str, paramdef: list[Union[Param, str]]) -> Optional[tuple[Param, OperatorType]]:
    casefoldqueryname = queryname.casefold()
    for param in paramdef:
        param = Param(name=param) if isinstance(param, str) else param
        operators = param.operators or ["="]
        for operator in operators:
            postfix = _operator_postfix_map[operator]
            if (param.name + postfix).casefold() == casefoldqueryname:
                return (param, operator)
    return None


def get_schema_for(model_ns: str, field: pa.Field) -> tuple[Union[type, Any], Any]:
    if pa.types.is_timestamp(field.type):
        return (Union[datetime.datetime, None], None)
    if pa.types.is_duration(field.type):
        return (Union[datetime.time, None], None)
    if pa.types.is_date(field.type):
        return (Union[datetime.date, None], None)
    if pa.types.is_time(field.type):
        return (Union[datetime.time, None], None)
    if pa.types.is_map(field.type):
        return (Union[dict, None], None)
    if pa.types.is_struct(field.type):
        st = field.type
        assert isinstance(st, pa.StructType)
        res = {st.field(ind).name: get_schema_for(model_ns, st.field(ind)) for ind in range(0, st.num_fields)}
        return (
            Union[
                create_model(model_ns + ("_" + field.name if field.name else ""), **res, __base__=TypeBaseModel),
                None,
            ],
            None,
        )
    if pa.types.is_list(field.type) or pa.types.is_large_list(field.type):
        if field.type.value_type is None:
            return (Union[List[Any], None], [])

        itemtype = cast(Any, get_schema_for(model_ns, field.type.value_type)[0])
        return (Union[List[itemtype], None], [])
    if pa.types.is_integer(field.type):
        return (Union[int, None], None)
    if pa.types.is_boolean(field.type):
        return (Union[bool, None], None)
    if pa.types.is_decimal(field.type) or pa.types.is_floating(field.type):
        return (Union[float, None], None)
    if pa.types.is_string(field.type) or pa.types.is_large_string(field.type):
        return (Union[str, None], None)
    if pa.types.is_union(field.type):
        types = [get_schema_for(model_ns, field.type.field(ind)) for ind in range(0, field.type.num_fields)]
        types.append(None)  # type: ignore
        return Union.__call__(*types)
    raise ValueError("Not supported")


def _get_datatype(schema: Optional[pa.Schema], name: str):
    if schema is None:
        return str
    try:
        field = schema.field(name)
        return get_schema_for("prm_" + name, field)[0]
    except KeyError as err:
        return str


def create_parameter_model(
    df: Optional[ResultData],
    name: str,
    params: Optional[Iterable[Union[Param, str]]],
    search: Optional[Iterable[SearchConfig]],
    apimethod: Literal["get", "post"],
):
    query_params = {}
    if params or search:
        for param in params or []:
            param = Param(name=param) if isinstance(param, str) else param
            operators = param.operators or ["="]
            schema = df.arrow_schema() if df else None
            if param.combi:
                if apimethod == "post":  # Only supported in POST Requets for now
                    query_params[param.name] = (
                        Union[List[dict[str, Any]], None],  # type: ignore
                        Query(default=param.real_default if not param.required else ...),
                    )
            else:
                realtype = (
                    _get_datatype(schema, param.name) if df is not None else str
                )  # well. no model, no real thing. just string
                for operator in operators:
                    postfix = _operator_postfix_map[operator]

                    if operator in ["in", "not in", "between", "not between"]:
                        if apimethod == "post":  # Only supported in POST Requets for now
                            query_params[param.name + postfix] = (
                                Union[List[realtype], List[dict[str, realtype]], None],  # type: ignore
                                Query(default=param.real_default if not param.required else ...),
                            )

                    else:
                        query_params[param.name + postfix] = (
                            realtype,
                            param.real_default if not param.required else ...,
                        )
        for sc in search or []:
            query_params[sc.name] = (
                Optional[str],
                None,
            )
        query_model = create_model(name + "Parameter", **query_params)
        return query_model
    return empty_model


class TypeBaseModel(BaseModel):
    class Config:
        orm_mode = True


def should_hide_colname(name: str):
    return name.startswith("_") or "_md5_prefix_" in name or "_xxhash64_prefix_" in name


def create_response_model(name: str, frame: ResultData) -> type[BaseModel]:
    schema = frame.arrow_schema()
    props = {k: get_schema_for(name, schema.field(k)) for k in schema.names if not should_hide_colname(k)}

    return create_model(name, **props, __base__=TypeBaseModel)



import inspect
from typing import Literal, Tuple, cast

from fastapi import APIRouter, Depends, Request

from bmsdna.lakeapi.core.config import BasicConfig, Configs
from bmsdna.lakeapi.core.log import get_logger


logger = get_logger(__name__)

all_lake_api_routers: list[Tuple[BasicConfig, Configs]] = []


def init_routes(configs: Configs, basic_config: BasicConfig):
    from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContext

    from bmsdna.lakeapi.core.endpoint import (
        get_response_model,
        create_detailed_meta_endpoint,
        create_config_endpoint,
        create_sql_endpoint,
    )

    async def get_username(req: Request):
        res = basic_config.username_retriever(req, basic_config, configs.users)
        if inspect.isawaitable(res):
            res = await res
        return res

    all_lake_api_routers.append((basic_config, configs))
    router = APIRouter()
    metadata = []

    with DuckDbExecutionContext() as context:
        for config in configs:
            methods = (
                cast(list[Literal["get", "post"]], [config.api_method])
                if isinstance(config.api_method, str)
                else config.api_method
            )
            try:
                from bmsdna.lakeapi.core.dataframe import Dataframe

                realdataframe = Dataframe(config.tag, config.name, config.dataframe, context, basic_config)
                if not realdataframe.file_exists():
                    logger.warning(
                        f"Could not get response type for f{config.route}. Path does not exist:{realdataframe.uri}"
                    )
                    metamodel = None
                else:
                    metamodel = realdataframe.get_df(endpoint="meta", partitions=None)
                schema = metamodel.arrow_schema() if metamodel else None
                metadata.append(
                    {
                        "name": config.name,
                        "tag": config.tag,
                        "route": config.route,
                        "methods": methods,
                        "file_type": config.dataframe.file_type,
                        "uri": config.dataframe.uri,
                        "version": config.version,
                        "schema": {n: str(schema.field(n).type) for n in schema.names} if schema else None,
                    }
                )

            except Exception as err:
                logger.warning(f"Could not get response type for f{config.route}. Error:{err}")
                metamodel = None

            response_model = get_response_model(config=config, metamodel=metamodel) if metamodel is not None else None
            create_detailed_meta_endpoint(metamodel=metamodel, config=config, router=router, basic_config=basic_config)
            for am in methods:
                create_config_endpoint(
                    apimethod=am,
                    config=config,
                    router=router,
                    response_model=response_model,
                    metamodel=metamodel,
                    basic_config=basic_config,
                    configs=configs,
                )

        @router.get(
            "/metadata",
            name="metadata",
        )
        async def get_metadata(username: str = Depends(get_username)):
            return metadata

        if basic_config.enable_sql_endpoint:
            create_sql_endpoint(
                router=router,
                basic_config=basic_config,
                configs=configs,
            )

        return router, get_username



import inspect
from pathlib import Path
from typing import List, Literal, Optional, Type, Union, cast

import duckdb
import polars as pl
import pyarrow as pa
import pypika
import pypika.queries as fn
from aiocache import Cache, cached
from aiocache.serializers import PickleSerializer
from deltalake import DeltaTable, Metadata, PyDeltaTableError
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    Header,
    HTTPException,
    Query,
    Request,
)
from pydantic import BaseModel
from pypika.queries import QueryBuilder

from bmsdna.lakeapi.context import get_context_by_engine
from bmsdna.lakeapi.context.df_base import ResultData
from bmsdna.lakeapi.core.config import BasicConfig, Config, Configs
from bmsdna.lakeapi.core.dataframe import (
    Dataframe,
    filter_df_based_on_params,
    filter_partitions_based_on_params,
)
from bmsdna.lakeapi.core.log import get_logger
from bmsdna.lakeapi.core.model import (
    create_parameter_model,
    create_response_model,
    should_hide_colname,
)
from bmsdna.lakeapi.core.response import create_response
from bmsdna.lakeapi.core.types import OutputFileType, Engines
from bmsdna.lakeapi.core.env import CACHE_EXPIRATION_TIME_SECONDS

cache = cached(ttl=CACHE_EXPIRATION_TIME_SECONDS, cache=Cache.MEMORY, serializer=PickleSerializer())

logger = get_logger(__name__)


def get_delta_folders(rootdir) -> dict[str, Path]:
    paths = [p.parent for p in Path(rootdir).rglob("*/_delta_log/")]
    return {p.parent.name + "_" + p.name: p for p in paths}


def register_duckdb_views(con: duckdb.DuckDBPyConnection, paths: dict[str, Path]):
    for name, path in paths.items():
        try:
            dt = DeltaTable(path.absolute())
            con.register(name, dt.to_pyarrow_dataset())
        except PyDeltaTableError as e:
            logger.warning(f"Could not register view for {name}. Error: {e}")
            pass


async def get_partitions(dataframe: Dataframe, params: BaseModel, config: Config) -> Optional[list]:
    parts = (
        await filter_partitions_based_on_params(
            DeltaTable(dataframe.uri).metadata(),
            params.dict(exclude_unset=True) if params else {},
            config.params or [],
        )
        if config.dataframe.file_type == "delta"
        else None
    )
    return parts


def remove_search(prm_dict: dict, config: Config):
    if not config.search or len(prm_dict) == 0:
        return prm_dict
    search_cols = [s.name.lower() for s in config.search]
    return {k: v for k, v in prm_dict.items() if k.lower() not in search_cols}


async def get_expr(columns: List[str], config: Config, params: BaseModel) -> Optional[pypika.Criterion]:
    expr = await filter_df_based_on_params(
        remove_search(params.dict(exclude_unset=True) if params else {}, config),
        config.params if config.params else [],
        columns,
    )
    return expr


def get_response_model(config: Config, metamodel: ResultData) -> Optional[Type[BaseModel]]:
    response_model: Optional[type[BaseModel]] = None
    try:
        response_model = create_response_model(config.tag + "_" + config.name, metamodel)
    except Exception as err:
        logger.warning(f"Could not get response type for f{config.route}. Error:{err}")
        response_model = None
    return response_model


def _to_dict(tblmeta: Optional[Metadata]):
    if not tblmeta:
        return None
    return {
        "name": tblmeta.name,
        "description": tblmeta.description,
        "partition_columns": tblmeta.partition_columns,
        "configuration": tblmeta.configuration,
    }


def create_detailed_meta_endpoint(
    metamodel: Optional[ResultData], config: Config, router: APIRouter, basic_config: BasicConfig
):
    route = config.route + "/metadata_detail"

    @router.get(
        route,
        tags=["metadata", "metadata:" + config.tag],
        operation_id=config.tag + "_" + config.name,
        name=config.name + "_metadata",
    )
    def get_detailed_metadata(req: Request):
        import json

        req.state.lake_api_basic_config = basic_config
        from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContext

        with DuckDbExecutionContext() as context:
            realdataframe = Dataframe(config.tag, config.name, config.dataframe, context, basic_config=basic_config)
            partition_columns = []
            partition_values = None
            delta_tbl = None
            df = realdataframe.get_df(None)
            if config.dataframe.file_type == "delta":
                delta_tbl = DeltaTable(realdataframe.uri)
                partition_columns = delta_tbl.metadata().partition_columns
                if len(partition_columns) > 0:
                    qb: QueryBuilder = (
                        df.query_builder().select(*[pypika.Field(c) for c in partition_columns]).distinct()
                    )
                    partition_values = context.execute_sql(qb).to_arrow_table().to_pylist()
            schema = df.arrow_schema()
            str_cols = [name for name in schema.names if pa.types.is_string(schema.field(name).type)]

            str_lengths_df = (
                (
                    context.execute_sql(
                        df.query_builder().select(
                            *[fn.Function("MAX", fn.Function("LEN", fn.Field(sc))).as_(sc) for sc in str_cols]
                        )
                    )
                    .to_arrow_table()
                    .to_pylist()
                )
                if len(str_cols) > 0
                else [{}]
            )
            str_lengths = str_lengths_df[0]

            def _recursive_get_type(t: Optional[pa.DataType]):
                if t is None:
                    return None
                return {
                    "type_str": str(t),
                    "base_type": str(t),  # legacy
                    "fields": [
                        {"name": f.name, "type": _recursive_get_type(f.type)}
                        for f in [t.field(find) for find in range(0, cast(pa.StructType, t).num_fields)]
                    ]
                    if pa.types.is_struct(t)
                    else None,
                    "inner": _recursive_get_type(t.value_type) if pa.types.is_list(t) else None,
                }

            schema = df.arrow_schema()
            return {
                "partition_values": partition_values,
                "partition_columns": partition_columns,
                "max_string_lengths": str_lengths,
                "schema": [{"name": n, "type": _recursive_get_type(schema.field(n).type)} for n in schema.names],
                "delta_meta": _to_dict(delta_tbl.metadata() if delta_tbl else None),
                "delta_schema": json.loads(delta_tbl.schema().to_json()) if delta_tbl else None,
                "parameters": config.params,
                "search": config.search,
            }


def exclude_cols(columns: List[str]) -> List[str]:
    columns = [c for c in columns if not should_hide_colname(c)]
    return columns


def create_config_endpoint(
    metamodel: Optional[ResultData],
    apimethod: Literal["get", "post"],
    config: Config,
    router: APIRouter,
    response_model: Type,
    basic_config: BasicConfig,
    configs: Configs,
):
    route = config.route

    query_model = create_parameter_model(
        metamodel, config.tag + "_" + config.name + "_" + apimethod, config.params, config.search, apimethod
    )

    async def get_username(req: Request):
        res = basic_config.username_retriever(req, basic_config, configs.users)
        if inspect.isawaitable(res):
            res = await res
        return res

    api_method_mapping = {
        "post": router.post(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
        "get": router.get(
            route,
            tags=[config.tag],
            response_model=response_model,
            operation_id=config.tag + "_" + config.name + "_" + apimethod,
            name=config.name,
        ),
    }

    api_method = api_method_mapping[apimethod]

    @api_method
    async def data(
        request: Request,
        params: query_model = (Depends() if apimethod == "get" else None),  # type: ignore
        Accept: Union[str, None] = Header(default=None),
        limit: Optional[int] = 100,
        offset: Optional[int] = 0,
        select: Union[str, None] = Query(title="$select", alias="$select", default=None, include_in_schema=False),
        distinct: bool = Query(title="$distinct", alias="$distinct", default=False, include_in_schema=False),
        engine: Engines = Query(title="$engine", alias="$engine", default="duckdb", include_in_schema=False),
        format: Optional[OutputFileType] = "json",
        username=Depends(get_username),
    ):  # type: ignore
        logger.info(f"{params.dict(exclude_unset=True) if params else None}Union[ ,  ]{request.url.path}")

        engine = engine or config.engine

        logger.info(f"Engine: {engine}")

        ExecutionContext = get_context_by_engine(engine)

        with ExecutionContext() as context:
            realdataframe = Dataframe(config.tag, config.name, config.dataframe, context, basic_config=basic_config)

            parts = await get_partitions(realdataframe, params, config)
            df = realdataframe.get_df(parts or None)

            expr = await get_expr(df.columns(), config, params)

            new_query = df.query_builder()
            new_query = new_query.where(expr) if expr is not None else new_query

            columns = exclude_cols(df.columns())

            if columns:
                new_query = new_query.select(*columns)
            if select:
                import pypika

                new_query = new_query.select(*[pypika.Field(s) for s in select.split(",")])
            if distinct:
                new_query = new_query.distinct()

            if not (limit == -1 and config.allow_get_all_pages):
                limit = 1000 if limit == -1 else limit
                new_query = new_query.offset(offset or 0).limit(limit)

            if config.search is not None and params is not None:
                search_dict = {c.name.lower(): c for c in config.search}
                searches = {
                    k: (v, search_dict[k.lower()])
                    for k, v in params.dict(exclude_unset=True).items()
                    if k.lower() in search_dict and v is not None and len(v) >= basic_config.min_search_length
                }
                if len(searches) > 0:
                    import pypika.queries
                    import pypika.terms

                    source_view = realdataframe.tablename
                    context.init_search(source_view, config.search)
                    score_sum = None
                    for search_key, (search_val, search_cfg) in searches.items():
                        score_sum = (
                            context.search_score_function(source_view, search_val, search_cfg, alias=None)
                            if score_sum is None
                            else score_sum
                            + context.search_score_function(source_view, search_val, search_cfg, alias=None)
                        )
                    assert score_sum is not None
                    new_query = new_query.select(score_sum.as_("search_score"))
                    new_query = new_query.where(pypika.terms.NotNullCriterion(pypika.queries.Field("search_score")))
                    new_query = new_query.orderby(pypika.queries.Field("search_score"), order=pypika.Order.desc)

            logger.info(f"Query: {new_query.get_sql()}")
            df2 = context.execute_sql(new_query)

            try:
                return await create_response(
                    username, request.url, format or request.headers["Accept"], df2, context, basic_config=basic_config
                )
            except Exception as err:
                logger.error("Error in creating response", exc_info=err)
                raise HTTPException(status_code=500)


duckcon: Optional[duckdb.DuckDBPyConnection] = None


def create_sql_endpoint(router: APIRouter, basic_config: BasicConfig, configs: Configs):
    paths = get_delta_folders(basic_config.data_path)

    @router.on_event("shutdown")
    async def shutdown_event():
        global duckcon
        if duckcon is not None:
            duckcon.close()

    async def get_username(req: Request):
        res = basic_config.username_retriever(req, basic_config, configs.users)
        if inspect.isawaitable(res):
            res = await res
        return res

    @router.get("/api/sql/tables", tags=["sql"], operation_id="get_sql_tables")
    async def get_sql_tables(
        request: Request,
        background_tasks: BackgroundTasks,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        username=Depends(get_username),
    ):
        try:
            return [table for table in paths]
        except Exception as err:
            raise HTTPException(status_code=500, detail=str(err))

    @router.post(
        "/api/sql",
        tags=["sql"],
        operation_id="post_sql_endpoint",
    )
    async def get_sql_post(
        request: Request,
        background_tasks: BackgroundTasks,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        username=Depends(get_username),
    ):
        try:
            body = await request.body()
            from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContextBase

            global duckcon
            if duckcon is None:
                duckcon = duckdb.connect()
                register_duckdb_views(duckcon, paths)
            context = DuckDbExecutionContextBase(duckcon)
            df = context.execute_sql(body.decode("utf-8"))

            return await create_response(
                username, request.url, format or request.headers["Accept"], df, context, basic_config=basic_config
            )
        except Exception as err:
            logger.error(err)
            raise HTTPException(status_code=500, detail=str(err))

    @router.get(
        "/api/sql",
        tags=["sql"],
        operation_id="get_sql_endpoint",
    )
    async def get_sql_get(
        request: Request,
        background_tasks: BackgroundTasks,
        sql: str,
        Accept: Union[str, None] = Header(default=None),
        format: Optional[OutputFileType] = "json",
        username=Depends(get_username),
    ):
        try:
            from bmsdna.lakeapi.context.df_duckdb import DuckDbExecutionContextBase

            global duckcon
            if duckcon is None:
                duckcon = duckdb.connect()
                register_duckdb_views(duckcon, paths)
            context = DuckDbExecutionContextBase(duckcon)

            df = context.execute_sql(sql)
            return await create_response(
                username, request.url, format or request.headers["Accept"], df, context, basic_config=basic_config
            )
        except Exception as err:
            logger.error(err)
            raise HTTPException(status_code=500, detail=str(err))
